------------------------------------------------------------------------------
--	FILE:	 Tilted_Axis.lua
--	AUTHOR:  Bob Thomas
--	PURPOSE: Global map script - A very large world, tilted on its side with 
--           its south pole always locked in to facing the sun.
------------------------------------------------------------------------------
--	Copyright (c) 2014 Firaxis Games, Inc. All rights reserved.
------------------------------------------------------------------------------

include("MapGenerator");
include("FractalWorld");
include("TerrainGenerator");
include("RiverGenerator");
include("FeatureGenerator");
include("MapmakerUtilities");

------------------------------------------------------------------------------
function GetMapScriptInfo()
	local world_age, temperature, rainfall, sea_level, resources = GetCoreMapOptions()
	local sea_level = {
		Name = "TXT_KEY_MAP_OPTION_SEA_LEVEL",
		Values = {
			{"TXT_KEY_MAP_OPTION_LOW"},
			{"TXT_KEY_MAP_OPTION_MEDIUM"},
			"TXT_KEY_MAP_OPTION_RANDOM",
		},
		DefaultValue = 1,
		SortPriority = -96,
	};
	return {
		Name = "TXT_KEY_MAP_TILTED_AXIS_NAME",
		Type = "TXT_KEY_MAP_TILTED_AXIS_TYPE",
		Description = "TXT_KEY_MAP_TILTED_AXIS_HELP",
		IsAdvancedMap = false,
		IconIndex = 10,
		CustomOptions = {world_age, rainfall, sea_level, resources,
			{
				Name = "TXT_KEY_MAP_OPTION_LANDMASS_TYPE",
				Values = {
					"TXT_KEY_MAP_OPTION_PANGAEA",
					"TXT_KEY_MAP_OPTION_LARGE_CONTINENTS",
					"TXT_KEY_MAP_OPTION_SMALL_CONTINENTS",
					"TXT_KEY_MAP_OPTION_RANDOM",
				},
				DefaultValue = 4,
				SortPriority = 1,
			},
		},
	}
end
------------------------------------------------------------------------------
function GetMapInitData(worldSize)
	-- This function can reset map grid sizes or world wrap settings.
	local worldsizes = {
		[GameInfo.Worlds.WORLDSIZE_DUEL.ID] = {52, 32},
		[GameInfo.Worlds.WORLDSIZE_TINY.ID] = {68, 44},
		[GameInfo.Worlds.WORLDSIZE_SMALL.ID] = {88, 56},
		[GameInfo.Worlds.WORLDSIZE_STANDARD.ID] = {108, 68},
		[GameInfo.Worlds.WORLDSIZE_LARGE.ID] = {118, 74},
		--[GameInfo.Worlds.WORLDSIZE_HUGE.ID] = {128, 80}
		}
	local grid_size = worldsizes[worldSize];
	--
	local world = GameInfo.Worlds[worldSize];
	if(world ~= nil) then
	return {
		Width = grid_size[1],
		Height = grid_size[2],
		WrapX = true,
	};      
     end
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
PangaeaFractalWorld = {};
------------------------------------------------------------------------------
function PangaeaFractalWorld.Create(fracXExp, fracYExp)
	local gridWidth, gridHeight = Map.GetGridSize();
	
	local data = {
		InitFractal = FractalWorld.InitFractal,
		ShiftPlotTypes = FractalWorld.ShiftPlotTypes,
		ShiftPlotTypesBy = FractalWorld.ShiftPlotTypesBy,
		DetermineXShift = FractalWorld.DetermineXShift,
		DetermineYShift = FractalWorld.DetermineYShift,
		GenerateCenterRift = FractalWorld.GenerateCenterRift,
		GeneratePlotTypes = PangaeaFractalWorld.GeneratePlotTypes,	-- Custom method
		
		-- Four methods have been crafted to add canyons to the maps. Method D occurs with plot generation. Methods A, B and C come with rivers.
		-- Method A: Fault lines where plates are pulling apart. Opposite of mountain range creation. These large canyons may be segmented.
		-- Method B: Weak spots in the planet's crust. Crust collapses in a pocket area due to high magma activity in the mantle. Creates isolated canyons.
		-- Method C: Plate collision creates mountains, then (geologically) rapid plate separation also creates a parallel canyon.
		-- Method D: Plate separation creates a canyon at fault line, then later plate collision closes most of the canyon and also creates new mountains.
		PlotCanBeCanyon = FractalWorld.PlotCanBeCanyon,
		HexAdjustment = FractalWorld.HexAdjustment,
		GenerateCanyonPlotCandidateList = FractalWorld.GenerateCanyonPlotCandidateList,
		MethodD = FractalWorld.MethodD,
		
		iFlags = Map.GetFractalFlags(),
		
		fracXExp = fracXExp,
		fracYExp = fracYExp,
		
		iNumPlotsX = gridWidth,
		iNumPlotsY = gridHeight,
		plotTypes = table.fill(PlotTypes.PLOT_OCEAN, gridWidth * gridHeight),

		-- Canyon variables
		canyon_plot_candidate_list = {},
		mountain_count = 0,
		extra_mountains = 0,
		adjustment = 0,
		
	};
		
	return data;
end	
------------------------------------------------------------------------------
function PangaeaFractalWorld:GeneratePlotTypes(args)
	if(args == nil) then args = {}; end
	
	local sea_level_low = 59;
	local sea_level_normal = 63;
	local sea_level_high = 68;
	local world_age_old = 2;
	local world_age_normal = 3;
	local world_age_new = 5;
	--
	local extra_mountains = 0;
	local grain_amount = 3;
	local adjust_plates = 1.3;
	local shift_plot_types = true;
	local tectonic_islands = true;
	local hills_ridge_flags = self.iFlags;
	local peaks_ridge_flags = self.iFlags;
	local has_center_rift = false;
	
	local sea_level = Map.GetCustomOption(3)
	if sea_level == 3 then
		sea_level = 1 + Map.Rand(2, "Random Sea Level - Lua");
	end
	local world_age = Map.GetCustomOption(1)
	if world_age == 4 then
		world_age = 1 + Map.Rand(3, "Random World Age - Lua");
	end

	-- Set Sea Level according to user selection.
	local water_percent = sea_level_normal;
	if sea_level == 1 then -- Low Sea Level
		water_percent = sea_level_low;
	else -- Normal Sea Level
	end

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = world_age_normal;
	if world_age == 3 then -- 5 Billion Years
		adjustment = world_age_old;
		adjust_plates = adjust_plates * 0.75;
	elseif world_age == 1 then -- 3 Billion Years
		adjustment = world_age_new;
		adjust_plates = adjust_plates * 1.5;
	else -- 4 Billion Years
	end
	-- Apply adjustment to hills and peaks settings.
	local hillsBottom1 = 28 - adjustment;
	local hillsTop1 = 28 + adjustment;
	local hillsBottom2 = 72 - adjustment;
	local hillsTop2 = 72 + adjustment;
	local hillsClumps = 1 + adjustment;
	local hillsNearMountains = 91 - (adjustment * 2) - extra_mountains;
	local mountains = 97 - adjustment - extra_mountains;

	-- Hills and Mountains handled differently according to map size - Bob
	local WorldSizeTypes = {};
	for row in GameInfo.Worlds() do
		WorldSizeTypes[row.Type] = row.ID;
	end
	local sizekey = Map.GetWorldSize();
	-- Fractal Grains
	local sizevalues = {
		[WorldSizeTypes.WORLDSIZE_DUEL]     = 3,
		[WorldSizeTypes.WORLDSIZE_TINY]     = 3,
		[WorldSizeTypes.WORLDSIZE_SMALL]    = 4,
		[WorldSizeTypes.WORLDSIZE_STANDARD] = 4,
		[WorldSizeTypes.WORLDSIZE_LARGE]    = 5,
		--[WorldSizeTypes.WORLDSIZE_HUGE]		= 5
	};
	local grain = sizevalues[sizekey] or 3;
	-- Tectonics Plate Counts
	local platevalues = {
		[WorldSizeTypes.WORLDSIZE_DUEL]		= 6,
		[WorldSizeTypes.WORLDSIZE_TINY]     = 9,
		[WorldSizeTypes.WORLDSIZE_SMALL]    = 12,
		[WorldSizeTypes.WORLDSIZE_STANDARD] = 18,
		[WorldSizeTypes.WORLDSIZE_LARGE]    = 24,
		--[WorldSizeTypes.WORLDSIZE_HUGE]     = 30
	};
	local numPlates = platevalues[sizekey] or 5;
	-- Add in any plate count modifications passed in from the map script. - Bob
	numPlates = numPlates * adjust_plates;

	-- Generate continental fractal layer and examine the largest landmass. Reject
	-- the result until the largest landmass occupies 84% or more of the total land.
	local done = false;
	local iAttempts = 0;
	local iWaterThreshold, biggest_area, iNumTotalLandTiles, iNumBiggestAreaTiles, iBiggestID;
	while done == false do
		local grain_dice = Map.Rand(7, "Continental Grain roll - LUA Pangaea");
		if grain_dice < 4 then
			grain_dice = 1;
		else
			grain_dice = 2;
		end
		local rift_dice = Map.Rand(3, "Rift Grain roll - LUA Pangaea");
		if rift_dice < 1 then
			rift_dice = -1;
		end
		
		self.continentsFrac = nil;
		self:InitFractal{continent_grain = grain_dice, rift_grain = rift_dice};
		iWaterThreshold = self.continentsFrac:GetHeight(water_percent);
		
		iNumTotalLandTiles = 0;
		for x = 0, self.iNumPlotsX - 1 do
			for y = 0, self.iNumPlotsY - 1 do
				local i = y * self.iNumPlotsX + x;
				local val = self.continentsFrac:GetHeight(x, y);
				if(val <= iWaterThreshold) then
					self.plotTypes[i] = PlotTypes.PLOT_OCEAN;
				else
					self.plotTypes[i] = PlotTypes.PLOT_LAND;
					iNumTotalLandTiles = iNumTotalLandTiles + 1;
				end
			end
		end

		SetPlotTypes(self.plotTypes);
		Map.RecalculateAreas();
		
		biggest_area = Map.FindBiggestArea(false);
		iNumBiggestAreaTiles = biggest_area:GetNumTiles();
		-- Now test the biggest landmass to see if it is large enough.
		if iNumBiggestAreaTiles >= iNumTotalLandTiles * 0.84 then
			done = true;
			iBiggestID = biggest_area:GetID();
		end
		iAttempts = iAttempts + 1;
		
		--[[ Printout for debug use only
		print("-"); print("--- Pangaea landmass generation, Attempt#", iAttempts, "---");
		print("- This attempt successful: ", done);
		print("- Total Land Plots in world:", iNumTotalLandTiles);
		print("- Land Plots belonging to biggest landmass:", iNumBiggestAreaTiles);
		print("- Percentage of land belonging to Pangaea: ", 100 * iNumBiggestAreaTiles / iNumTotalLandTiles);
		print("- Continent Grain for this attempt: ", grain_dice);
		print("- Rift Grain for this attempt: ", rift_dice);
		print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
		print(".");
		]]--
	end
	
	-- Generate fractals to govern hills and mountains
	self.hillsFrac = Fractal.Create(self.iNumPlotsX, self.iNumPlotsY, grain, self.iFlags, self.fracXExp, self.fracYExp);
	self.mountainsFrac = Fractal.Create(self.iNumPlotsX, self.iNumPlotsY, grain, self.iFlags, self.fracXExp, self.fracYExp);
	self.hillsFrac:BuildRidges(numPlates, hills_ridge_flags, 1, 2);
	self.mountainsFrac:BuildRidges((numPlates * 2) / 3, peaks_ridge_flags, 6, 1);
	-- Get height values
	local iHillsBottom1 = self.hillsFrac:GetHeight(hillsBottom1);
	local iHillsTop1 = self.hillsFrac:GetHeight(hillsTop1);
	local iHillsBottom2 = self.hillsFrac:GetHeight(hillsBottom2);
	local iHillsTop2 = self.hillsFrac:GetHeight(hillsTop2);
	local iHillsClumps = self.mountainsFrac:GetHeight(hillsClumps);
	local iHillsNearMountains = self.mountainsFrac:GetHeight(hillsNearMountains);
	local iMountainThreshold = self.mountainsFrac:GetHeight(mountains);
	local iPassThreshold = self.hillsFrac:GetHeight(hillsNearMountains);
	-- Get height values for tectonic islands
	local iMountain100 = self.mountainsFrac:GetHeight(100);
	local iMountain99 = self.mountainsFrac:GetHeight(99);
	local iMountain97 = self.mountainsFrac:GetHeight(97);
	local iMountain95 = self.mountainsFrac:GetHeight(95);

	-- Because we haven't yet shifted the plot types, we will not be able to take advantage 
	-- of having water and flatland plots already set. We still have to generate all data
	-- for hills and mountains, too, then shift everything, then set plots one more time.
	for x = 0, self.iNumPlotsX - 1 do
		for y = 0, self.iNumPlotsY - 1 do
		
			local i = y * self.iNumPlotsX + x;
			local val = self.continentsFrac:GetHeight(x, y);
			local mountainVal = self.mountainsFrac:GetHeight(x, y);
			local hillVal = self.hillsFrac:GetHeight(x, y);
	
			if(val <= iWaterThreshold) then
				self.plotTypes[i] = PlotTypes.PLOT_OCEAN;
				
				if tectonic_islands then -- Build islands in oceans along tectonic ridge lines - Brian
					if (mountainVal == iMountain100) then -- Isolated peak in the ocean
						self.plotTypes[i] = PlotTypes.PLOT_MOUNTAIN;
					elseif (mountainVal == iMountain99) then
						self.plotTypes[i] = PlotTypes.PLOT_HILLS;
					elseif (mountainVal == iMountain97) or (mountainVal == iMountain95) then
						self.plotTypes[i] = PlotTypes.PLOT_LAND;
					end
				end
					
			else
				if (mountainVal >= iMountainThreshold) then
					if (hillVal >= iPassThreshold) then -- Mountain Pass though the ridgeline - Brian
						self.plotTypes[i] = PlotTypes.PLOT_HILLS;
					else -- Mountain
						self.plotTypes[i] = PlotTypes.PLOT_MOUNTAIN;
					end
				elseif (mountainVal >= iHillsNearMountains) then
					self.plotTypes[i] = PlotTypes.PLOT_HILLS; -- Foot hills - Bob
				else
					if ((hillVal >= iHillsBottom1 and hillVal <= iHillsTop1) or (hillVal >= iHillsBottom2 and hillVal <= iHillsTop2)) then
						self.plotTypes[i] = PlotTypes.PLOT_HILLS;
					else
						self.plotTypes[i] = PlotTypes.PLOT_LAND;
					end
				end
			end
		end
	end

	self:ShiftPlotTypes();
	
	-- Now shift everything toward the south pole, to tilt the continent toward the warm climate.
	local iStartRow, iNumRowsToShift;
	local bFoundPangaea, bDoShift = false, false;
	-- Shift South
	for y = 1, self.iNumPlotsY - 2 do
		for x = 0, self.iNumPlotsX - 1 do
			local i = y * self.iNumPlotsX + x;
			if self.plotTypes[i] == PlotTypes.PLOT_HILLS or self.plotTypes[i] == PlotTypes.PLOT_LAND then
				local plot = Map.GetPlot(x, y);
				local iAreaID = plot:GetArea();
				if iAreaID == iBiggestID then
					bFoundPangaea = true;
					iStartRow = y - 1;
					if iStartRow > 3 then -- Enough rows of water space to do a shift.
						bDoShift = true;
					end
					break
				end
			end
		end
		-- Check to see if we've found the Pangaea.
		if bFoundPangaea == true then
			break
		end
	end
	if bDoShift == true then
		local iRowsDifference = iStartRow - 1;
		local iRowsInPlay = math.floor(iRowsDifference * 0.7);
		local iRowsBase = math.ceil(iRowsDifference * 0.3);
		local rows_dice = Map.Rand(iRowsInPlay, "Number of Rows to Shift - LUA Pangaea");
		local iNumRows = math.min(iRowsDifference - 1, iRowsBase + rows_dice);
		local iNumEvenRows = 2 * math.floor(iNumRows / 2); -- MUST be an even number or we risk breaking a 1-tile isthmus and splitting the Pangaea.
		local iNumRowsToShift = math.max(2, iNumEvenRows);
		--print("-"); print("Shifting lands southward by this many plots: ", iNumRowsToShift); print("-");
		-- Process from bottom up.
		for y = 0, (self.iNumPlotsY - 1) - iNumRowsToShift do
			for x = 0, self.iNumPlotsX - 1 do
				local sourcePlotIndex = (y + iNumRowsToShift) * self.iNumPlotsX + x + 1;
				local destPlotIndex = y * self.iNumPlotsX + x + 1;
				self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
			end
		end
		for y = self.iNumPlotsY - iNumRowsToShift, self.iNumPlotsY - 1 do
			for x = 0, self.iNumPlotsX - 1 do
				local i = y * self.iNumPlotsX + x + 1;
				self.plotTypes[i] = PlotTypes.PLOT_OCEAN;
			end
		end
	end

	-- Canyon generation handled independent from other plot generation, for easier customization.
	-- Individual components of canyon generation can be overridden without having to override the entire plot generator.
	-- Method D is called here. Methods A, B and C are called in the RiverGenerator.
	self:GenerateCanyonPlotCandidateList()
	self:MethodD()

	return self.plotTypes;
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
ContinentsFractalWorld = {};
------------------------------------------------------------------------------
function ContinentsFractalWorld.Create(fracXExp, fracYExp)
	local gridWidth, gridHeight = Map.GetGridSize();
	
	local data = {
		InitFractal = FractalWorld.InitFractal,
		ShiftPlotTypes = FractalWorld.ShiftPlotTypes,
		ShiftPlotTypesBy = FractalWorld.ShiftPlotTypesBy,
		DetermineXShift = FractalWorld.DetermineXShift,
		DetermineYShift = FractalWorld.DetermineYShift,
		GenerateCenterRift = FractalWorld.GenerateCenterRift,
		GeneratePlotTypes = ContinentsFractalWorld.GeneratePlotTypes,	-- Custom method
		
		-- Four methods have been crafted to add canyons to the maps. Method D occurs with plot generation. Methods A, B and C come with rivers.
		-- Method A: Fault lines where plates are pulling apart. Opposite of mountain range creation. These large canyons may be segmented.
		-- Method B: Weak spots in the planet's crust. Crust collapses in a pocket area due to high magma activity in the mantle. Creates isolated canyons.
		-- Method C: Plate collision creates mountains, then (geologically) rapid plate separation also creates a parallel canyon.
		-- Method D: Plate separation creates a canyon at fault line, then later plate collision closes most of the canyon and also creates new mountains.
		PlotCanBeCanyon = ContinentsFractalWorld.PlotCanBeCanyon,
		HexAdjustment = FractalWorld.HexAdjustment,
		GenerateCanyonPlotCandidateList = ContinentsFractalWorld.GenerateCanyonPlotCandidateList,
		MethodD = ContinentsFractalWorld.MethodD,
		ContsPlotCanBeCanyon = ContinentsFractalWorld.ContsPlotCanBeCanyon,
		ContsGenerateCanyonPlotCandidateList = ContinentsFractalWorld.ContsGenerateCanyonPlotCandidateList,
		ContsMethodD = ContinentsFractalWorld.ContsMethodD,
		
		iFlags = Map.GetFractalFlags(),
		
		fracXExp = fracXExp,
		fracYExp = fracYExp,
		
		iNumPlotsX = gridWidth,
		iNumPlotsY = gridHeight,
		plotTypes = table.fill(PlotTypes.PLOT_OCEAN, gridWidth * gridHeight),

		-- Canyon variables
		canyon_plot_candidate_list = {},
		mountain_count = 0,
		extra_mountains = 0,
		adjustment = 0,

	};
		
	return data;
end	
------------------------------------------------------------------------------
function ContinentsFractalWorld:ContsPlotCanBeCanyon(plot)
	if plot:IsWater() or plot:IsMountain() or plot:IsCanyon() then
		return false
	else
		return true
	end
end
------------------------------------------------------------------------------
function ContinentsFractalWorld:ContsGenerateCanyonPlotCandidateList()
	for x = 0, self.iNumPlotsX - 1 do
		for y = 1, self.iNumPlotsY - 2 do -- Never putting a canyon in top or bottom row of map.
			local plot = Map.GetPlot(x, y);
			local i = y * self.iNumPlotsX + x + 1;
			if plot:IsWater() then
				self.canyon_plot_candidate_list[i] = false;
			elseif plot:IsMountain() then
				self.mountain_count = self.mountain_count + 1;
				self.canyon_plot_candidate_list[i] = false;
			else
				self.canyon_plot_candidate_list[i] = true;
			end
		end
	end
end
------------------------------------------------------------------------------
function ContinentsFractalWorld:ContsMethodD()
	-- This method seeks to simulate areas where plates pulled apart first, then slammed back 
	-- together, first creating fault line canyons, then smashing most of them and making mountains.
	--
	-- The actual method is to identify hills plots of a specific fractal elevation, in
	-- relationship to mountain generation, and turn some of these and one or two adjacent 
	-- plots into canyons, always two or three plots in size. The net effect should be some
	-- cases of two or three plot canyons in the shadow of mountain ranges.
	print("Mountain count = ", self.mountain_count);
	local max_type_d_canyon_plots = math.ceil(self.mountain_count / 3);
	print("Max 'Type D' canyon plots = ", max_type_d_canyon_plots)
	local canyon_seed = 92 - self.extra_mountains - self.adjustment;
	local HeightTypeD = self.mountainsFrac:GetHeight(canyon_seed);
	local type_d_seed_list = {};
	local num_seed_plots = 0;
	local canyons_placed = 0;
	local firstRingYIsEven = {{0, 1}, {1, 0}, {0, -1}, {-1, -1}, {-1, 0}, {-1, 1}};
	local firstRingYIsOdd = {{1, 1}, {1, 0}, {1, -1}, {0, -1}, {-1, 0}, {0, 1}};
	
	-- Generate seed list.
	for x = 0, self.iNumPlotsX - 1 do
		for y = 0, self.iNumPlotsY - 1 do
			local plot = Map.GetPlot(x, y);
			if self:ContsPlotCanBeCanyon(plot) then
				local canyonVal = self.mountainsFrac:GetHeight(x, y);
				if canyonVal == HeightTypeD then
					table.insert(type_d_seed_list, {x, y});
					num_seed_plots = num_seed_plots + 1;
				end
			end
		end
	end
	local shuffled_seed_list;
	
	-- Apply Method D.
	if num_seed_plots > 0 then
		shuffled_seed_list = GetShuffledCopyOfTable(type_d_seed_list);
		for loop = 1, num_seed_plots do
			local x, y = shuffled_seed_list[loop][1], shuffled_seed_list[loop][2];
			local plot = Map.GetPlot(x, y)
			plot:SetPlotType(PlotTypes.PLOT_CANYON, false, false);
			canyons_placed = canyons_placed + 1;
			local i = y * self.iNumPlotsX + x + 1;
			self.canyon_plot_candidate_list[i] = false;
			print("Placing Canyon center via Method D at plot: ", x, y);
			
			-- Try to add adjacent canyon plots.
			local rand = Map.Rand(5, "Extra canyon plots - Lua FractalWorld");
			local num_extra = 1;
			if rand > 1 then
				num_extra = 2;
			end
			print("Attempting to add ", num_extra, " wing plots to this canyon.")
			local wing_plot_candidate_list = {};
			local randomized_first_ring_adjustments = nil;
			local isEvenY = true;
			if y / 2 > math.floor(y / 2) then
				isEvenY = false;
			end
			if isEvenY then
				randomized_first_ring_adjustments = GetShuffledCopyOfTable(firstRingYIsEven);
			else
				randomized_first_ring_adjustments = GetShuffledCopyOfTable(firstRingYIsOdd);
			end
			for attempt = 1, 6 do
				local plot_adjustments = randomized_first_ring_adjustments[attempt];
				local searchX, searchY = self:HexAdjustment(x, y, plot_adjustments)
				local search_i = searchY * self.iNumPlotsX + searchX + 1;
				-- Make sure the search plot is not off the grid.
				if search_i > 0 and search_i < self.iNumPlotsX * self.iNumPlotsY then -- OK, it's in play.
					if self.canyon_plot_candidate_list[search_i] == true then
						table.insert(wing_plot_candidate_list, {searchX, searchY});
					end
				end
			end
			local iNumWingCandidates = table.maxn(wing_plot_candidate_list);
			if iNumWingCandidates >= 1 then
				local iNumWings = math.min(num_extra, iNumWingCandidates)
				-- Create Canyon "wing" plots.
				for wing_loop = 1, iNumWings do
					local wing_x, wing_y = wing_plot_candidate_list[wing_loop][1], wing_plot_candidate_list[wing_loop][2];
					local wing_plot = Map.GetPlot(wing_x, wing_y)
					wing_plot:SetPlotType(PlotTypes.PLOT_CANYON, false, false);
					canyons_placed = canyons_placed + 1;
					local wing_i = wing_y - 1 * self.iNumPlotsX + wing_x + 1;
					self.canyon_plot_candidate_list[wing_i] = false;
					print("Placing Canyon wing via Method D at plot: ", wing_x, wing_y);
				end
			else
				print("No wings available for canyon at ", x, y);
			end

			if canyons_placed >= max_type_d_canyon_plots then
				print(canyons_placed, "canyon plots placed using Method D.");
				break
			elseif loop == num_seed_plots then
				print(canyons_placed, "canyon plots placed using Method D. Less than target: ran out of candidates.");
			end
		end
	else
		print("No eligible plots for Method D of Canyon generation.")
	end
end
------------------------------------------------------------------------------
function ContinentsFractalWorld:GeneratePlotTypes(args)
	if(args == nil) then args = {}; end
	
	local sea_level_low = 63;
	local sea_level_normal = 67;
	local sea_level_high = 71;
	local world_age_old = 2;
	local world_age_normal = 3;
	local world_age_new = 5;
	--
	local extra_mountains = 0;
	local grain_amount = 3;
	local adjust_plates = 1.0;
	local shift_plot_types = true;
	local tectonic_islands = false;
	local hills_ridge_flags = self.iFlags;
	local peaks_ridge_flags = self.iFlags;
	local has_center_rift = true;
	
	local sea_level = Map.GetCustomOption(3)
	if sea_level == 3 then
		sea_level = 1 + Map.Rand(2, "Random Sea Level - Lua");
	end
	local world_age = Map.GetCustomOption(1)
	if world_age == 4 then
		world_age = 1 + Map.Rand(3, "Random World Age - Lua");
	end

	-- Set Sea Level according to user selection.
	local water_percent = sea_level_normal;
	if sea_level == 1 then -- Low Sea Level
		water_percent = sea_level_low
	else -- Normal Sea Level
	end

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = world_age_normal;
	if world_age == 3 then -- 5 Billion Years
		adjustment = world_age_old;
		adjust_plates = adjust_plates * 0.75;
	elseif world_age == 1 then -- 3 Billion Years
		adjustment = world_age_new;
		adjust_plates = adjust_plates * 1.5;
	else -- 4 Billion Years
	end
	-- Apply adjustment to hills and peaks settings.
	local hillsBottom1 = 28 - adjustment;
	local hillsTop1 = 28 + adjustment;
	local hillsBottom2 = 72 - adjustment;
	local hillsTop2 = 72 + adjustment;
	local hillsClumps = 1 + adjustment;
	local hillsNearMountains = 91 - (adjustment * 2) - extra_mountains;
	local mountains = 97 - adjustment - extra_mountains;

	-- Hills and Mountains handled differently according to map size
	local WorldSizeTypes = {};
	for row in GameInfo.Worlds() do
		WorldSizeTypes[row.Type] = row.ID;
	end
	local sizekey = Map.GetWorldSize();
	-- Fractal Grains
	local sizevalues = {
		[WorldSizeTypes.WORLDSIZE_DUEL]     = 3,
		[WorldSizeTypes.WORLDSIZE_TINY]     = 3,
		[WorldSizeTypes.WORLDSIZE_SMALL]    = 4,
		[WorldSizeTypes.WORLDSIZE_STANDARD] = 4,
		[WorldSizeTypes.WORLDSIZE_LARGE]    = 5,
		--[WorldSizeTypes.WORLDSIZE_HUGE]		= 5
	};
	local grain = sizevalues[sizekey] or 3;
	-- Tectonics Plate Counts
	local platevalues = {
		[WorldSizeTypes.WORLDSIZE_DUEL]		= 6,
		[WorldSizeTypes.WORLDSIZE_TINY]     = 9,
		[WorldSizeTypes.WORLDSIZE_SMALL]    = 12,
		[WorldSizeTypes.WORLDSIZE_STANDARD] = 18,
		[WorldSizeTypes.WORLDSIZE_LARGE]    = 24,
		--[WorldSizeTypes.WORLDSIZE_HUGE]     = 30
	};
	local numPlates = platevalues[sizekey] or 5;
	-- Add in any plate count modifications passed in from the map script.
	numPlates = numPlates * adjust_plates;

	-- Generate continental fractal layer and examine the largest landmass. Reject
	-- the result until the largest landmass occupies 58% or less of the total land.
	local done = false;
	local iAttempts = 0;
	local iWaterThreshold, biggest_area, iNumTotalLandTiles, iNumBiggestAreaTiles, iBiggestID;
	while done == false do
		local grain_dice = Map.Rand(7, "Continental Grain roll - LUA Continents");
		if grain_dice < 4 then
			grain_dice = 2;
		else
			grain_dice = 1;
		end
		local rift_dice = Map.Rand(3, "Rift Grain roll - LUA Continents");
		if rift_dice < 1 then
			rift_dice = -1;
		end
		
		self.continentsFrac = nil;
		self:InitFractal{continent_grain = grain_dice, rift_grain = rift_dice};
		iWaterThreshold = self.continentsFrac:GetHeight(water_percent);
		
		iNumTotalLandTiles = 0;
		for x = 0, self.iNumPlotsX - 1 do
			for y = 0, self.iNumPlotsY - 1 do
				local i = y * self.iNumPlotsX + x;
				local val = self.continentsFrac:GetHeight(x, y);
				if(val <= iWaterThreshold) then
					self.plotTypes[i] = PlotTypes.PLOT_OCEAN;
				else
					self.plotTypes[i] = PlotTypes.PLOT_LAND;
					iNumTotalLandTiles = iNumTotalLandTiles + 1;
				end
			end
		end

		self:ShiftPlotTypes();
		self:GenerateCenterRift()

		SetPlotTypes(self.plotTypes);
		Map.RecalculateAreas();
		
		biggest_area = Map.FindBiggestArea(false);
		iNumBiggestAreaTiles = biggest_area:GetNumTiles();
		-- Now test the biggest landmass to see if it is large enough.
		if iNumBiggestAreaTiles <= iNumTotalLandTiles * 0.58 then
			done = true;
			iBiggestID = biggest_area:GetID();
		end
		iAttempts = iAttempts + 1;
		
		--[[ Printout for debug use only
		print("-"); print("--- Continents landmass generation, Attempt#", iAttempts, "---");
		print("- This attempt successful: ", done);
		print("- Total Land Plots in world:", iNumTotalLandTiles);
		print("- Land Plots belonging to biggest landmass:", iNumBiggestAreaTiles);
		print("- Percentage of land belonging to biggest: ", 100 * iNumBiggestAreaTiles / iNumTotalLandTiles);
		print("- Continent Grain for this attempt: ", grain_dice);
		print("- Rift Grain for this attempt: ", rift_dice);
		print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
		print(".");
		]]--
	end
	
	-- Generate fractals to govern hills and mountains
	self.hillsFrac = Fractal.Create(self.iNumPlotsX, self.iNumPlotsY, grain, self.iFlags, self.fracXExp, self.fracYExp);
	self.mountainsFrac = Fractal.Create(self.iNumPlotsX, self.iNumPlotsY, grain, self.iFlags, self.fracXExp, self.fracYExp);
	self.hillsFrac:BuildRidges(numPlates, hills_ridge_flags, 1, 2);
	self.mountainsFrac:BuildRidges((numPlates * 2) / 3, peaks_ridge_flags, 6, 1);
	-- Get height values
	local iHillsBottom1 = self.hillsFrac:GetHeight(hillsBottom1);
	local iHillsTop1 = self.hillsFrac:GetHeight(hillsTop1);
	local iHillsBottom2 = self.hillsFrac:GetHeight(hillsBottom2);
	local iHillsTop2 = self.hillsFrac:GetHeight(hillsTop2);
	local iHillsClumps = self.mountainsFrac:GetHeight(hillsClumps);
	local iHillsNearMountains = self.mountainsFrac:GetHeight(hillsNearMountains);
	local iMountainThreshold = self.mountainsFrac:GetHeight(mountains);
	local iPassThreshold = self.hillsFrac:GetHeight(hillsNearMountains);
	
	-- Set Hills and Mountains
	for x = 0, self.iNumPlotsX - 1 do
		for y = 0, self.iNumPlotsY - 1 do
			local plot = Map.GetPlot(x, y);
			local mountainVal = self.mountainsFrac:GetHeight(x, y);
			local hillVal = self.hillsFrac:GetHeight(x, y);
	
			if plot:GetPlotType() ~= PlotTypes.PLOT_OCEAN then
				if (mountainVal >= iMountainThreshold) then
					if (hillVal >= iPassThreshold) then -- Mountain Pass though the ridgeline
						plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
					else -- Mountain
						plot:SetPlotType(PlotTypes.PLOT_MOUNTAIN, false, false);
					end
				elseif (mountainVal >= iHillsNearMountains) then
					plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
				elseif ((hillVal >= iHillsBottom1 and hillVal <= iHillsTop1) or (hillVal >= iHillsBottom2 and hillVal <= iHillsTop2)) then
					plot:SetPlotType(PlotTypes.PLOT_HILLS, false, false);
				end
			end
		end
	end

	-- Canyon generation handled independent from other plot generation, for easier customization.
	-- Individual components of canyon generation can be overridden without having to override the entire plot generator.
	-- Method D is called here. Methods A, B and C are called in the RiverGenerator.
	self:ContsGenerateCanyonPlotCandidateList()
	self:ContsMethodD()

end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
function GeneratePlotTypes()
	print("Generating Plot Types (Lua Tilted Axis) ...");
	
	-- Obtain landmass type selected by user.
	userInputLandmass = Map.GetCustomOption(5) -- GLOBAL variable
	if userInputLandmass == 4 then -- Random
		userInputLandmass = 1 + Map.Rand(3, "Tilted Axis Random Landmass Type - Lua");
	end

	-- Implement landmass type.
	if userInputLandmass == 2 then -- Continents
		local fractal_world = ContinentsFractalWorld.Create();
		fractal_world:GeneratePlotTypes();
	
		GenerateCoasts();
		
	elseif userInputLandmass == 3 then -- Small Continents
		local sea_level = Map.GetCustomOption(3)
		if sea_level == 3 then
			sea_level = 1 + Map.Rand(2, "Random Sea Level - Lua");
		end
		local world_age = Map.GetCustomOption(1)
		if world_age == 4 then
			world_age = 1 + Map.Rand(3, "Random World Age - Lua");
		end

		local fractal_world = FractalWorld.Create();
		fractal_world:InitFractal{
			continent_grain = 3};

		local args = {
			sea_level = sea_level,
			world_age = world_age,
			sea_level_low = 64,
			sea_level_normal = 70,
			sea_level_high = 75,
			extra_mountains = 2,
			adjust_plates = 1.5,
			tectonic_islands = true,
			}
		local plotTypes = fractal_world:GeneratePlotTypes(args);

		-- Now shift everything toward the south pole, to tilt the land toward the warm climate.
		local iW, iH = Map.GetGridSize()
		local iStartRow, iNumRowsToShift = 4, 4;
		-- Process from bottom up.
		for y = 0, (iH - 1) - iNumRowsToShift do
			for x = 0, iW - 1 do
				local sourcePlotIndex = (y + iNumRowsToShift) * iW + x + 1;
				local destPlotIndex = y * iW + x + 1;
				plotTypes[destPlotIndex] = plotTypes[sourcePlotIndex]
			end
		end
		for y = iH - iNumRowsToShift, iH - 1 do
			for x = 0, iW - 1 do
				local i = y * iW + x + 1;
				plotTypes[i] = PlotTypes.PLOT_OCEAN;
			end
		end
	
		SetPlotTypes(plotTypes);
		GenerateCoasts();

	else -- Pangaea
		local fractal_world = PangaeaFractalWorld.Create();
		local plotTypes = fractal_world:GeneratePlotTypes();
	
		SetPlotTypes(plotTypes);
		GenerateCoasts();

	end
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
function TerrainGenerator:GetLatitudeAtPlot(iX, iY)
	-- Custom lat for Tilted Axis.
	local lat = 2 * iY/self.iHeight;
	lat = lat + (128 - self.variation:GetHeight(iX, iY))/(255.0 * 5.0);
	lat = math.clamp(lat, 0, 2);
	return lat
end
----------------------------------------------------------------------------------
function TerrainGenerator:GenerateTerrainAtPlot(iX,iY)
	local lat = self:GetLatitudeAtPlot(iX,iY);

	local plot = Map.GetPlot(iX, iY);
	if (plot:IsWater()) then
		local val = plot:GetTerrainType();
		if val == TerrainTypes.NO_TERRAIN then -- Error handling.
			val = self.terrainGrass;
			plot:SetPlotType(PlotTypes.PLOT_LAND, false, false);
		end
		return val;	 
	end
	
	local terrainVal = self.terrainGrass;

	if lat >= 1 then
		terrainVal = self.terrainSnow;
	elseif lat >= 0.8 then
		terrainVal = self.terrainTundra;
	elseif lat < 0.35 then
		terrainVal = self.terrainDesert;
	elseif lat <= 0.5 then
		terrainVal = self.terrainGrass;
	else
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = self.terrainDesert;
		elseif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = self.terrainPlains;
		end
	end
	
	-- Error handling.
	if (terrainVal == TerrainTypes.NO_TERRAIN) then
		return plot:GetTerrainType();
	end

	return terrainVal;
end
------------------------------------------------------------------------------
function GenerateTerrain()
	print("Generating Terrain (Lua Tilted Axis) ...");
	local args = {iDesertPercent = 11, iPlainsPercent = 35,	};
	local terraingen = TerrainGenerator.Create(args);

	local terrainTypes = terraingen:GenerateTerrain()
		
	SetTerrainTypes(terrainTypes);
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
function RiverGenerator:GetCapsForMethodC()
	-- Set up caps for number of formations and plots based on world size.
	if userInputLandmass == 3 then -- Small Continents, reduce Method C parameters.
		local worldsizes = {
			[GameInfo.Worlds.WORLDSIZE_DUEL.ID] = {2, 4, 2, 0},
			[GameInfo.Worlds.WORLDSIZE_TINY.ID] = {3, 7, 2, 1},
			[GameInfo.Worlds.WORLDSIZE_SMALL.ID] = {5, 12, 2, 2},
			[GameInfo.Worlds.WORLDSIZE_STANDARD.ID] = {6, 18, 2, 3},
			[GameInfo.Worlds.WORLDSIZE_LARGE.ID] = {7, 28, 3, 3},
			--[GameInfo.Worlds.WORLDSIZE_HUGE.ID] = {9, 42, 4, 4},
			};
		local caps_list = worldsizes[Map.GetWorldSize()];
		local max_lines, max_plots, base_length, extension_range = caps_list[1], caps_list[2], caps_list[3], caps_list[4];
		return max_lines, max_plots, base_length, extension_range;
	else -- Regular parameters.
		local worldsizes = {
			[GameInfo.Worlds.WORLDSIZE_DUEL.ID] = {2, 6, 4, 2},
			[GameInfo.Worlds.WORLDSIZE_TINY.ID] = {3, 10, 5, 2},
			[GameInfo.Worlds.WORLDSIZE_SMALL.ID] = {4, 14, 6, 2},
			[GameInfo.Worlds.WORLDSIZE_STANDARD.ID] = {5, 21, 7, 3},
			[GameInfo.Worlds.WORLDSIZE_LARGE.ID] = {6, 32, 8, 3},
			--[GameInfo.Worlds.WORLDSIZE_HUGE.ID] = {8, 50, 10, 4},
			};
		local caps_list = worldsizes[Map.GetWorldSize()];
		local max_lines, max_plots, base_length, extension_range = caps_list[1], caps_list[2], caps_list[3], caps_list[4];
		return max_lines, max_plots, base_length, extension_range;
	end
end
------------------------------------------------------------------------------
function AddRivers()
	print("Generating Rivers, Canyons, and Lakes. (Lua Tilted Axis) ...");

	local args = {};
	local rivergen = RiverGenerator.Create(args);
	
	rivergen:Generate();
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
function FeatureGenerator:GetLatitudeAtPlot(iX, iY)
	local lat = 2 * iY / self.iGridH;
	return lat
end
------------------------------------------------------------------------------
function FeatureGenerator:AddFeaturesAtPlot(iX, iY)
	-- adds any appropriate features at the plot (iX, iY) where (0,0) is in the SW
	local lat = self:GetLatitudeAtPlot(iX, iY);
	local plot = Map.GetPlot(iX, iY);

	-- Determine Wildness of this plot.
	local terrain_type = plot:GetTerrainType()
	if terrain_type == self.terrainDesert then
		local plot_type = plot:GetPlotType()
		if plot_type == PlotTypes.PLOT_LAND or plot_type == PlotTypes.PLOT_HILLS then
			local iWildVal = self.miasma:GetHeight(iX, iY)
			if iWildVal >= self.iWildAreaHeight then
				plot:SetWildness(20) -- Designate as a "core" wild plot in a desert Wild Area.
			else
				plot:SetWildness(21) -- Designate as a "periphery" wild plot in a desert Wild Area.
			end
		end
	elseif lat < 1 then
		local plot_type = plot:GetPlotType()
		if plot_type == PlotTypes.PLOT_LAND or plot_type == PlotTypes.PLOT_HILLS then
			if self.forestclumps:GetHeight(iX, iY) >= self.iClumpLevel then -- Wild plot.
				local iWildVal = self.miasma:GetHeight(iX, iY)
				if iWildVal >= self.iWildAreaHeight then
					if terrain_type == self.terrainTundra then
						plot:SetWildness(30) -- Designate as a "core" wild plot in a tundra Wild Area.
					elseif terrain_type == self.terrainGrass or terrain_type == self.terrainPlains then
						plot:SetWildness(10) -- Designate as a "core" wild plot in a forest Wild Area.
					end
				else
					if terrain_type == self.terrainTundra then
						plot:SetWildness(31) -- Designate as a "periphery" wild plot in a tundra Wild Area.
					elseif terrain_type == self.terrainGrass or terrain_type == self.terrainPlains then
						plot:SetWildness(11) -- Designate as a "periphery" wild plot in a forest Wild Area.
					end
				end
			end
		end
	end

	if plot:CanHaveFeature(self.featureFloodPlains) then
		-- All desert plots along river are set to flood plains.
		plot:SetFeatureType(self.featureFloodPlains, -1)
	end
	
	if (plot:GetFeatureType() == FeatureTypes.NO_FEATURE) then
		self:AddIceAtPlot(plot, iX, iY, lat);
	end

	if (plot:GetFeatureType() == FeatureTypes.NO_FEATURE) then
		self:AddMarshAtPlot(plot, iX, iY, lat);
	end
		
	if (plot:GetFeatureType() == FeatureTypes.NO_FEATURE) then
		self:AddForestsAtPlot(plot, iX, iY, lat);
	end
		
	if (plot:GetFeatureType() == FeatureTypes.NO_FEATURE) then
		self:AddJunglesAtPlot(plot, iX, iY, lat);
	end
	
	self:DetermineWildness(plot, iX, iY, lat);
	
end
------------------------------------------------------------------------------
function FeatureGenerator:AddIceAtPlot(plot, iX, iY, lat)
	if(plot:CanHaveFeature(self.featureIce)) then
		local iW, iH = Map.GetGridSize()
		if iY >= iH / 2 then
			plot:SetFeatureType(self.featureIce, -1);
		else
			local rand = Map.Rand(100, "Add Ice Lua") / 100;

			if(rand < 8 * (lat - 0.975)) then
				plot:SetFeatureType(self.featureIce, -1);
			elseif(rand < 4 * (lat - 0.85)) then
				plot:SetFeatureType(self.featureIce, -1);
			end
		end
	end
end
------------------------------------------------------------------------------
function FeatureGenerator:AddJunglesAtPlot(plot, iX, iY, lat)
	local lat = (lat - 0.425) * 1.3;
	-- Jungles are not placed in BE. This function has been repurposed to place a
	-- combination of marsh, clear and forest features in place of where jungles went.
	local jungle_height = self.jungles:GetHeight(iX, iY);
	if jungle_height <= self.iJungleTop and jungle_height >= self.iJungleBottom + (self.iJungleRange * lat) then
		if(plot:CanHaveFeature(self.featureJungle)) then
			local repurpose_height = self.repurpose:GetHeight(iX, iY);
			if repurpose_height > self.iTop then
				local plotType = plot:GetPlotType()
				if plotType ~= PlotTypes.PLOT_HILLS then
					plot:SetFeatureType(self.featureMarsh, -1);
				end
			elseif repurpose_height < self.iBottom then
				plot:SetFeatureType(self.featureForest, -1);
			else -- Leave this plot clear.
			end
		end
	end
end
------------------------------------------------------------------------------
function FeatureGenerator:AdjustTerrainTypes()
	local width = self.iGridW - 1;
	local height = self.iGridH - 1;
	
	for y = 0, math.floor(height * 0.48) do -- For Tilted Axis, only adjust on light side of the planet.
		for x = 0, width do
			local plot = Map.GetPlot(x, y);
			
			if (plot:GetFeatureType() == self.featureJungle) then
				plot:SetTerrainType(self.terrainPlains, false, true)  -- These flags are for recalc of areas and rebuild of graphics. No need to recalc from any of these changes.		
			elseif (plot:IsRiver()) then
				local terrainType = plot:GetTerrainType();
				if (terrainType == self.terrainTundra) then
					plot:SetTerrainType(self.terrainPlains, false, true)
				elseif (terrainType == self.terrainIce) then
					plot:SetTerrainType(self.terrainTundra, false, true)					
				end
			end
		end
	end
end
------------------------------------------------------------------------------
function FeatureGenerator:AddAtolls()
	-- Do nothing.
end
------------------------------------------------------------------------------
function AddFeatures()
	print("Adding Features (Lua Tilted Axis) ...");
	
	-- Get Rainfall setting input by user.
	local rain = Map.GetCustomOption(2)
	if rain == 4 then
		rain = 1 + Map.Rand(3, "Random Rainfall - Lua");
	end
	
	local args = {rainfall = rain}
	local featuregen = FeatureGenerator.Create(args);

	-- False parameter removes mountains from coastlines.
	featuregen:AddFeatures(false);
end
------------------------------------------------------------------------------

------------------------------------------------------------------------------
function AssignStartingPlots:__CustomInit()
	-- Using this override to slip in Geothermal placement in the ice, along
	-- with associated terrain softening custom to Tilted Axis.
	--
	self.softened_ice_plots, self.river_ice_plots, self.next_to_mountain_ice_plots, self.next_to_canyon_ice_plots = {}, {}, {}, {};
	self.next_to_desert_wild_plots, self.next_to_forest_wild_plots, self.next_to_tundra_wild_plots, self.next_to_softened_ice_plots = {}, {}, {}, {};
	local iW, iH = Map.GetGridSize()
	local candidate_plots, chosen_plots = {}, {};
	local y_bottom, y_top = math.floor(iH * 0.51), math.floor(iH * 0.75); -- Candidate range: "Latitude 1.02 to 1.5, on scale of 0 to 2."
	-- Within the desired vertical range, identify candidate plots.
	for y = y_bottom, y_top do
		for x = 0, iW - 1 do
			local plot = Map.GetPlot(x, y)
			local plotType = plot:GetPlotType()
			if plotType == PlotTypes.PLOT_LAND then
				if plot:GetResourceType() == -1 then
					if not plot:IsCoastalLand() then
						table.insert(candidate_plots, {x, y});
					end
				end
			end
		end
	end
	local iNumCandPlots = table.maxn(candidate_plots);
	if iNumCandPlots < 1 then
		print("-"); print("*** No candidates found for Geothermal centers in the ice!"); print("-"); 
		return
	end
	-- Shuffle the plot list. Determine the desired number of plots to assign.
	local shuffled_candidate_plots = GetShuffledCopyOfTable(candidate_plots)
	local iNumGeothermalCenters = math.ceil(iNumCandPlots / 30);
	
	print("-"); print("** Placing Large Geothermal deposits in the ice.");
	-- Create the Geothermal centers.
	for loop = 1, iNumGeothermalCenters do
		print("-");
		local x, y = shuffled_candidate_plots[loop][1], shuffled_candidate_plots[loop][2];
		local plot = Map.GetPlot(x, y)
		if plot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW then -- Proceed. (Have to recheck for this each time, as plots tied to a previous center have been converted to tundra.)
			local diceroll = Map.Rand(6, "Geothermal deposit size - Lua Tilted Axis");
			local size = 4 + diceroll;
			table.insert(chosen_plots, {x, y});
			plot:SetTerrainType(TerrainTypes.TERRAIN_TUNDRA, false, true);
			plot:SetResourceType(self.geothermal_ID, size);
			self.amounts_of_resources_placed[self.geothermal_ID + 1] = self.amounts_of_resources_placed[self.geothermal_ID + 1] + size;
			print("- Placed Geothermal of size " .. size .. " at plot " .. x .. ", " .. y);
			-- Now choose periphery plots to convert to tundra as well. Choose same amount as the size of the deposit.
			local remain_to_place = size;
			local first_ring_candidate_plots, second_ring_candidate_plots = {}, {};
			local plot_adjustments_table = self.firstRingYIsEven;
			if y / 2 > math.floor(y / 2) then
				plot_adjustments_table = self.firstRingYIsOdd;
			end
			for attempt = 1, 6 do
				local plot_adjustments = plot_adjustments_table[attempt];
				local searchX, searchY = self:ApplyHexAdjustment(x, y, plot_adjustments)
				local search_plot = Map.GetPlot(searchX, searchY);
				if search_plot ~= nil then
					if not (search_plot:IsWater() or search_plot:IsMountain() or search_plot:IsCanyon()) then -- Hills, or Flat: proceed.
						if search_plot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW then
							table.insert(first_ring_candidate_plots, {searchX, searchY});
						end
					end
				end
			end
			local shuffled_first_ring = GetShuffledCopyOfTable(first_ring_candidate_plots)
			local target = table.maxn(shuffled_first_ring);
			-- Assign first ring plots to be softened from Snow to Tundra, and add these plots to the Geothermal Center plot list.
			if target > 0 then
				for loop = 1, target do
					local geo_x, geo_y = shuffled_first_ring[loop][1], shuffled_first_ring[loop][2];
					local geo_plot = Map.GetPlot(geo_x, geo_y)
					geo_plot:SetTerrainType(TerrainTypes.TERRAIN_TUNDRA, false, true);
					local i = geo_y * iW + geo_x + 1;
					table.insert(self.softened_ice_plots, i);
					print("- Added first-ring plot " .. geo_x .. ", " .. geo_y .. " to Geothermal Center.");
					remain_to_place = remain_to_place - 1;
					if remain_to_place == 0 then
						break
					end
				end
			end
			if remain_to_place > 0 then -- More to do. Proceed to second ring.
				local sr_plot_adjustments_table = self.secondRingYIsEven;
				if y / 2 > math.floor(y / 2) then
					sr_plot_adjustments_table = self.secondRingYIsOdd;
				end
				for sr_attempt = 1, 12 do
					local sr_plot_adjustments = sr_plot_adjustments_table[sr_attempt];
					local searchX, searchY = self:ApplyHexAdjustment(x, y, sr_plot_adjustments)
					print("Attempt = " .. sr_attempt .. ". Search Plot = " .. searchX .. ", " .. searchY);
					local search_plot = Map.GetPlot(searchX, searchY);
					if search_plot ~= nil then
						if not (search_plot:IsWater() or search_plot:IsMountain() or search_plot:IsCanyon()) then -- Hills, or Flat: proceed.
							if search_plot:GetTerrainType() == TerrainTypes.TERRAIN_SNOW then
								-- Plot itself is eligible, but we must first determine if it's next to an existing first ring softened plot.
								-- We want the area to be contiguous, no isolated plots by themselves in the snow.
								local viable_candidate = false;
								local scan_plot_adjustments_table = self.firstRingYIsEven;
								if searchY / 2 > math.floor(searchY / 2) then
									scan_plot_adjustments_table = self.firstRingYIsOdd;
								end
								for scan_attempt = 1, 6 do
									local scan_plot_adjustments = scan_plot_adjustments_table[scan_attempt];
									local scan_x, scan_y = self:ApplyHexAdjustment(searchX, searchY, scan_plot_adjustments)
									local scan_plot = Map.GetPlot(scan_x, scan_y);
									if not (scan_plot:IsWater() or scan_plot:IsMountain() or scan_plot:IsCanyon()) then 
										if scan_plot:GetTerrainType() == TerrainTypes.TERRAIN_TUNDRA then
											print("> Found softened plot!");
											viable_candidate = true;
										end
									end
								end
								if viable_candidate == true then
									table.insert(second_ring_candidate_plots, {searchX, searchY});
								end
							end
						end
					end
				end
				local shuffled_second_ring = GetShuffledCopyOfTable(second_ring_candidate_plots)
				local second_target = table.maxn(shuffled_second_ring);
				print("* Candidate plots for second ring: ", second_target);
				if second_target > 0 then
					local chosen_target = math.min(second_target, remain_to_place);
					for loop = 1, chosen_target do
						local geo_x, geo_y = shuffled_second_ring[loop][1], shuffled_second_ring[loop][2];
						local geo_plot = Map.GetPlot(geo_x, geo_y)
						geo_plot:SetTerrainType(TerrainTypes.TERRAIN_TUNDRA, false, true);
						local i = geo_y * iW + geo_x + 1;
						table.insert(self.softened_ice_plots, i);
						print("- Added second-ring plot " .. geo_x .. ", " .. geo_y .. " to Geothermal Center.");
						remain_to_place = remain_to_place - 1;
						if remain_to_place == 0 then
							break
						end
					end
				end
			end
		end
	end
	print("-");
end	
------------------------------------------------------------------------------
function AssignStartingPlots:GenerateGlobalResourcePlotLists()
	-- This function generates all global plot lists needed for resource distribution.
	local iW, iH = Map.GetGridSize();
	local temp_coast_next_to_land_list, temp_marsh_list, temp_flood_plains_list = {}, {}, {};
	local temp_hills_open_list, temp_hills_covered_list, temp_hills_jungle_list = {}, {}, {};
	local temp_hills_forest_list, temp_jungle_flat_list, temp_forest_flat_list = {}, {}, {};
	local temp_desert_flat_no_feature, temp_plains_flat_no_feature, temp_dry_grass_flat_no_feature = {}, {}, {};
	local temp_fresh_water_grass_flat_no_feature, temp_tundra_flat_including_forests, temp_forest_flat_that_are_not_tundra = {}, {}, {};
	--
	-- Lists for BE.
	local temp_next_to_canyon_list, temp_coastal_land_list, temp_next_to_mountain_list = {}, {}, {};
	local temp_river_list, temp_wild_area_list, temp_wild_area_flatlands_list, temp_next_to_wild_area_list = {}, {}, {}, {};
	local temp_wild_desert_list, temp_wild_tundra_list, temp_wild_ocean_list = {}, {}, {};
	local temp_large_xenomass_list, temp_large_floatstone_list, temp_large_firaxite_list = {}, {}, {};
	local temp_small_xenomass_list, temp_small_floatstone_list, temp_small_firaxite_list = {}, {}, {};
	local temp_loose_xenomass_list, temp_loose_floatstone_list, temp_loose_firaxite_list = {}, {}, {};
	--
	local temp_next_to_softened_ice_plots, temp_river_ice_plots, temp_next_to_mountain_ice_plots, temp_next_to_canyon_ice_plots = {}, {}, {}, {};
	local temp_next_to_desert_wild_plots, temp_next_to_forest_wild_plots, temp_next_to_tundra_wild_plots = {}, {}, {};
	--
	local iW, iH = Map.GetGridSize();
	local temp_hills_list, temp_coast_list, temp_grass_flat_no_feature = {}, {}, {};
	local temp_tundra_flat_no_feature, temp_snow_flat_list, temp_land_list = {}, {}, {};
	local temp_deer_list, temp_desert_wheat_list, temp_banana_list = {}, {}, {};
	--
	for y = 0, iH - 1 do
		for x = 0, iW - 1 do
			local i = y * iW + x + 1; -- Lua tables/lists/arrays start at 1, not 0 like C++ or Python
			local plot = Map.GetPlot(x, y)
			-- Check if plot has a civ start, CS start, or Natural Wonder
			if self.playerCollisionData[i] == true then
				-- Do not process this plot!
			elseif plot:GetResourceType(-1) ~= -1 then
				-- Plot has a resource already, do not include it.
			else
				-- Process this plot for inclusion in the plot lists.
				local plotType = plot:GetPlotType()
				local terrainType = plot:GetTerrainType()
				local featureType = plot:GetFeatureType()
				local lat = 2 * y / iH;
				if plotType == PlotTypes.PLOT_MOUNTAIN or plotType == PlotTypes.PLOT_CANYON then
					self.barren_plots = self.barren_plots + 1;
				elseif plotType == PlotTypes.PLOT_OCEAN then
					if featureType ~= self.feature_atoll then
						if featureType == FeatureTypes.FEATURE_ICE then
							self.barren_plots = self.barren_plots + 1;
						elseif plot:IsLake() then
							self.barren_plots = self.barren_plots + 1;
						else
							if is_ocean_wild then
								table.insert(temp_wild_ocean_list, i);
							end
							if terrainType == TerrainTypes.TERRAIN_COAST then
								table.insert(temp_coast_list, i);
								if plot:IsAdjacentToLand() then
									table.insert(temp_coast_next_to_land_list, i);
								end
							else
								self.barren_plots = self.barren_plots + 1;
							end
						end
					end
				else -- Plot is hills or flat land.
					if lat >= 1 then -- Plot is hills or flat land on the planet's dark side. Needs special handling.
						if plot:IsRiver() then
							if plotType == PlotTypes.PLOT_LAND then
								table.insert(temp_river_ice_plots, i);
							end
						end
						-- Check all adjacent plots for canyons, mountains, and Geothermal Centers.
						local plot_adjustments_table = self.firstRingYIsEven;
						local isEvenY = true;
						if y / 2 > math.floor(y / 2) then
							plot_adjustments_table = self.firstRingYIsOdd;
						end
						for attempt = 1, 6 do
							local plot_adjustments = plot_adjustments_table[attempt];
							local searchX, searchY = self:ApplyHexAdjustment(x, y, plot_adjustments)
							local search_plot = Map.GetPlot(searchX, searchY);
							if search_plot ~= nil then
								if search_plot:IsCanyon() then
									table.insert(temp_next_to_canyon_ice_plots, i);
								elseif search_plot:IsMountain() then
									table.insert(temp_next_to_mountain_ice_plots, i);
								elseif terrain_type == TerrainTypes.TERRAIN_SNOW then -- Checking the core plot here not the search plot, on this line.
									if search_plot:GetTerrainType() == TerrainTypes.TERRAIN_TUNDRA then
										table.insert(temp_next_to_softened_ice_plots, i);
									end
								end
							end
						end
					else -- Plot is hills or flat land. Check for adjacencies as well as the plot's own traits.
						-- Check Wildness for the Key (Affinity-enabling) resources.
						local iWildValue = plot:GetWildness()
						local is_any_wild, is_forest_wild, is_desert_wild, is_tundra_wild = true, false, false, false;
						if iWildValue == 11 then -- Periphery forest wild plot.
							table.insert(temp_small_xenomass_list, i);
							is_forest_wild = true;
						elseif iWildValue == 21 then -- Periphery desert wild plot.
							table.insert(temp_small_floatstone_list, i);
							is_desert_wild = true;
						elseif iWildValue == 31 then -- Periphery tundra wild plot.
							table.insert(temp_small_firaxite_list, i);
							is_tundra_wild = true;
						elseif iWildValue == 10 then -- Core forest wild plot.
							table.insert(temp_large_xenomass_list, i);
							is_forest_wild = true;
						elseif iWildValue == 20 then -- Core desert wild plot.
							table.insert(temp_large_floatstone_list, i);
							is_desert_wild = true;
						elseif iWildValue == 30 then -- Core tundra wild plot.
							table.insert(temp_large_firaxite_list, i);
							is_tundra_wild = true;
						else
							is_any_wild = false;
						end
						-- On with the plot evaluation.
						if plot:IsCoastalLand() and terrainType ~= TerrainTypes.TERRAIN_SNOW then
							table.insert(temp_coastal_land_list, i);
						elseif plot:IsRiver() then
							if plotType == PlotTypes.PLOT_LAND then
								table.insert(temp_river_list, i);
								if is_any_wild == false then
									table.insert(temp_loose_xenomass_list, i);
								end
							end
						end
						-- Check all adjacent plots for canyons, mountains, and wild area.
						local plot_adjustments_table = self.firstRingYIsEven;
						local isEvenY = true;
						if y / 2 > math.floor(y / 2) then
							plot_adjustments_table = self.firstRingYIsOdd;
						end
						for attempt = 1, 6 do
							local plot_adjustments = plot_adjustments_table[attempt];
							local searchX, searchY = self:ApplyHexAdjustment(x, y, plot_adjustments)
							local search_plot = Map.GetPlot(searchX, searchY);
							if search_plot ~= nil then
								if search_plot:IsCanyon() then
									table.insert(temp_next_to_canyon_list, i);
									if is_any_wild == false then
										table.insert(temp_loose_floatstone_list, i);
									end
								elseif search_plot:IsMountain() then
									table.insert(temp_next_to_mountain_list, i);
									if is_any_wild == false then
										table.insert(temp_loose_floatstone_list, i);
									end
								elseif is_any_wild == false then
									local search_plot_wild_value = search_plot:GetWildness()
									if search_plot_wild_value >= 10 and search_plot_wild_value < 40 then
										table.insert(temp_next_to_wild_area_list, i);
									end
								end
							end
						end
					
						if plotType == PlotTypes.PLOT_HILLS then
							if terrainType == TerrainTypes.TERRAIN_SNOW then
								if is_tundra_wild then
									table.insert(temp_wild_tundra_list, i);
								end
							else
								table.insert(temp_hills_list, i);
								if featureType == FeatureTypes.NO_FEATURE then
									table.insert(temp_hills_open_list, i);
									if is_desert_wild then
										table.insert(temp_wild_desert_list, i);
									elseif is_tundra_wild then
										table.insert(temp_wild_tundra_list, i);
									elseif is_forest_wild == false then
										table.insert(temp_loose_firaxite_list, i);
									end
								elseif featureType == FeatureTypes.FEATURE_FOREST then
									if is_forest_wild then
										table.insert(temp_wild_area_list, i);
									elseif is_tundra_wild then
										table.insert(temp_wild_tundra_list, i);
									end
									table.insert(temp_hills_forest_list, i);
									table.insert(temp_hills_covered_list, i);
									if terrainType == TerrainTypes.TERRAIN_TUNDRA then
										table.insert(temp_deer_list, i);
									end
								end
							end
						elseif featureType == FeatureTypes.FEATURE_MARSH then
							table.insert(temp_marsh_list, i);
							if is_any_wild == false then
								table.insert(temp_loose_xenomass_list, i);
							end
						elseif featureType == FeatureTypes.FEATURE_FLOOD_PLAINS then
							table.insert(temp_flood_plains_list, i);
							table.insert(temp_desert_wheat_list, i);
							if is_desert_wild then
								table.insert(temp_wild_desert_list, i);
							end
						elseif plotType == PlotTypes.PLOT_LAND then
							table.insert(temp_land_list, i);
							if featureType == FeatureTypes.FEATURE_FOREST then
								if is_forest_wild then
									table.insert(temp_wild_area_list, i);
									table.insert(temp_wild_area_flatlands_list, i);
								end
								table.insert(temp_forest_flat_list, i);
								if terrainType == TerrainTypes.TERRAIN_TUNDRA then
									if is_tundra_wild then
										table.insert(temp_wild_tundra_list, i);
									end
									table.insert(temp_deer_list, i);
									table.insert(temp_tundra_flat_including_forests, i);
								else
									table.insert(temp_forest_flat_that_are_not_tundra, i);
									if is_any_wild == false then
										table.insert(temp_loose_xenomass_list, i);
									end
								end
							elseif featureType == FeatureTypes.NO_FEATURE then
								if terrainType == TerrainTypes.TERRAIN_SNOW then
									if is_tundra_wild then
										table.insert(temp_wild_tundra_list, i);
									else
										table.insert(temp_loose_floatstone_list, i);
									end
									table.insert(temp_snow_flat_list, i);
								elseif terrainType == TerrainTypes.TERRAIN_TUNDRA then
									if is_tundra_wild then
										table.insert(temp_wild_tundra_list, i);
									end
									table.insert(temp_tundra_flat_no_feature, i);
									table.insert(temp_tundra_flat_including_forests, i);
								elseif terrainType == TerrainTypes.TERRAIN_DESERT then
									if is_desert_wild then
										table.insert(temp_wild_desert_list, i);
									else
										table.insert(temp_loose_firaxite_list, i);
									end
									table.insert(temp_desert_flat_no_feature, i);
									if plot:IsFreshWater() then
										table.insert(temp_desert_wheat_list, i);
									end
								elseif terrainType == TerrainTypes.TERRAIN_PLAINS then
									table.insert(temp_plains_flat_no_feature, i);
									--if plot:IsFreshWater() == false then
									--end
								elseif terrainType == TerrainTypes.TERRAIN_GRASS then
									table.insert(temp_grass_flat_no_feature, i);
									if plot:IsFreshWater() then
										table.insert(temp_fresh_water_grass_flat_no_feature, i);
									else
										table.insert(temp_dry_grass_flat_no_feature, i);
									end
								else
									self.barren_plots = self.barren_plots + 1;
									table.remove(temp_land_list);
								end
							else
								self.barren_plots = self.barren_plots + 1;
								table.remove(temp_land_list);
							end
						else
							self.barren_plots = self.barren_plots + 1;
						end
					end
				end
			end
		end
	end
	
	print("Finished scanning plots for global plot lists."); print("-");
	
	-- Scramble and record the lists.
	self.coast_next_to_land_list = GetShuffledCopyOfTable(temp_coast_next_to_land_list)
	self.marsh_list = GetShuffledCopyOfTable(temp_marsh_list)
	self.flood_plains_list = GetShuffledCopyOfTable(temp_flood_plains_list)
	self.hills_open_list = GetShuffledCopyOfTable(temp_hills_open_list)
	self.hills_covered_list = GetShuffledCopyOfTable(temp_hills_covered_list)
	self.hills_jungle_list = GetShuffledCopyOfTable(temp_hills_jungle_list)
	self.hills_forest_list = GetShuffledCopyOfTable(temp_hills_forest_list)
	self.jungle_flat_list = GetShuffledCopyOfTable(temp_jungle_flat_list)
	self.forest_flat_list = GetShuffledCopyOfTable(temp_forest_flat_list)
	self.desert_flat_no_feature = GetShuffledCopyOfTable(temp_desert_flat_no_feature)
	self.plains_flat_no_feature = GetShuffledCopyOfTable(temp_plains_flat_no_feature)
	self.dry_grass_flat_no_feature = GetShuffledCopyOfTable(temp_dry_grass_flat_no_feature)
	self.fresh_water_grass_flat_no_feature = GetShuffledCopyOfTable(temp_fresh_water_grass_flat_no_feature)
	self.tundra_flat_including_forests = GetShuffledCopyOfTable(temp_tundra_flat_including_forests)
	self.forest_flat_that_are_not_tundra = GetShuffledCopyOfTable(temp_forest_flat_that_are_not_tundra)
	--
	self.grass_flat_no_feature = GetShuffledCopyOfTable(temp_grass_flat_no_feature)
	self.tundra_flat_no_feature = GetShuffledCopyOfTable(temp_tundra_flat_no_feature)
	self.snow_flat_list = GetShuffledCopyOfTable(temp_snow_flat_list)
	self.hills_list = GetShuffledCopyOfTable(temp_hills_list)
	self.land_list = GetShuffledCopyOfTable(temp_land_list)
	self.coast_list = GetShuffledCopyOfTable(temp_coast_list)
	self.extra_deer_list = GetShuffledCopyOfTable(temp_deer_list)
	self.desert_wheat_list = GetShuffledCopyOfTable(temp_desert_wheat_list)
	self.banana_list = GetShuffledCopyOfTable(temp_banana_list)
	--
	self.next_to_canyon_list = GetShuffledCopyOfTable(temp_next_to_canyon_list)
	self.coastal_land_list = GetShuffledCopyOfTable(temp_coastal_land_list)
	self.next_to_mountain_list = GetShuffledCopyOfTable(temp_next_to_mountain_list)
	self.river_list = GetShuffledCopyOfTable(temp_river_list)
	self.wild_area_list = GetShuffledCopyOfTable(temp_wild_area_list)
	self.wild_area_flatlands_list = GetShuffledCopyOfTable(temp_wild_area_flatlands_list)
	self.next_to_wild_area_list = GetShuffledCopyOfTable(temp_next_to_wild_area_list)
	self.wild_desert_list = GetShuffledCopyOfTable(temp_wild_desert_list)
	self.wild_tundra_list = GetShuffledCopyOfTable(temp_wild_tundra_list)
	self.wild_ocean_list = GetShuffledCopyOfTable(temp_wild_ocean_list)
	self.large_xenomass_list = GetShuffledCopyOfTable(temp_large_xenomass_list)
	self.large_firaxite_list = GetShuffledCopyOfTable(temp_large_firaxite_list)
	self.large_floatstone_list = GetShuffledCopyOfTable(temp_large_floatstone_list)
	self.small_xenomass_list = GetShuffledCopyOfTable(temp_small_xenomass_list)
	self.small_firaxite_list = GetShuffledCopyOfTable(temp_small_firaxite_list)
	self.small_floatstone_list = GetShuffledCopyOfTable(temp_small_floatstone_list)
	self.loose_xenomass_list = GetShuffledCopyOfTable(temp_loose_xenomass_list)
	self.loose_firaxite_list = GetShuffledCopyOfTable(temp_loose_firaxite_list)
	self.loose_floatstone_list = GetShuffledCopyOfTable(temp_loose_floatstone_list)
	--
	local temp_softened_ice_plots = self.softened_ice_plots; -- This list was formed during the __Custom_Init() function.
	self.softened_ice_plots = GetShuffledCopyOfTable(temp_softened_ice_plots)
	self.river_ice_plots = GetShuffledCopyOfTable(temp_river_ice_plots)
	self.next_to_mountain_ice_plots = GetShuffledCopyOfTable(temp_next_to_mountain_ice_plots)
	self.next_to_canyon_ice_plots = GetShuffledCopyOfTable(temp_next_to_canyon_ice_plots)
	self.next_to_desert_wild_plots = GetShuffledCopyOfTable(temp_next_to_desert_wild_plots)
	self.next_to_forest_wild_plots = GetShuffledCopyOfTable(temp_next_to_forest_wild_plots)
	self.next_to_tundra_wild_plots = GetShuffledCopyOfTable(temp_next_to_tundra_wild_plots)
	self.next_to_softened_ice_plots = GetShuffledCopyOfTable(temp_next_to_softened_ice_plots)
	--
	local iNumXeno = table.maxn(self.large_xenomass_list);
	local iNumFirax = table.maxn(self.softened_ice_plots);
	local iNumFloat = table.maxn(self.large_floatstone_list);
	local iNumSmallXeno = table.maxn(self.small_xenomass_list);
	local iNumSmallFirax = table.maxn(self.small_firaxite_list);
	local iNumSmallFloat = table.maxn(self.small_floatstone_list);
	local iNumLooseXeno = table.maxn(self.loose_xenomass_list);
	local iNumLooseFirax = table.maxn(self.loose_firaxite_list);
	local iNumLooseFloat = table.maxn(self.loose_floatstone_list);
	print("-"); print("************************"); print("*");
	print("* Number of candidate plots for large deposits of Xenomass: ", iNumXeno);
	print("* Number of candidate plots for large deposits of Firaxite: ", iNumFirax);
	print("* Number of candidate plots for large deposits of Float Stone: ", iNumFloat); print("*");
	print("* Number of candidate plots for small deposits of Xenomass: ", iNumSmallXeno);
	print("* Number of candidate plots for small deposits of Firaxite: ", iNumSmallFirax);
	print("* Number of candidate plots for small deposits of Float Stone: ", iNumSmallFloat); print("*");
	print("* Number of candidate plots for Xenomass outside wild areas: ", iNumLooseXeno);
	print("* Number of candidate plots for Firaxite outside wild areas: ", iNumLooseFirax);
	print("* Number of candidate plots for Float Stone outside wild areas: ", iNumLooseFloat);
	print("*"); print("************************"); print("-");
	--
	-- Set up the Global Luxury Plot Lists matrix, with indices synched to GetIndicesForLuxuryType()
	self.global_luxury_plot_lists = {
	self.coast_next_to_land_list,				-- 1
	self.marsh_list,							-- 2
	self.flood_plains_list,						-- 3
	self.hills_open_list,						-- 4
	self.hills_covered_list,					-- 5
	self.hills_jungle_list,						-- 6
	self.hills_forest_list,						-- 7
	self.jungle_flat_list,						-- 8
	self.forest_flat_list,						-- 9
	self.desert_flat_no_feature,				-- 10
	self.plains_flat_no_feature,				-- 11
	self.dry_grass_flat_no_feature,				-- 12
	self.fresh_water_grass_flat_no_feature,		-- 13
	self.tundra_flat_including_forests,			-- 14
	self.forest_flat_that_are_not_tundra,		-- 15
	};

end
------------------------------------------------------------------------------
function AssignStartingPlots:GetKeyStrategicsTargetValues()
	-- This is a BE function in support of PlaceKeyStrategics.
	-- Override this function to alter the target amounts. It's possible to set each resource's targets individually.
	local worldsizes = {
		[GameInfo.Worlds.WORLDSIZE_DUEL.ID] = {2, 3},
		[GameInfo.Worlds.WORLDSIZE_TINY.ID] = {3, 4},
		[GameInfo.Worlds.WORLDSIZE_SMALL.ID] = {4, 5},
		[GameInfo.Worlds.WORLDSIZE_STANDARD.ID] = {5, 7},
		[GameInfo.Worlds.WORLDSIZE_LARGE.ID] = {6, 8},
		--[GameInfo.Worlds.WORLDSIZE_HUGE.ID] = {7, 9}
		}
	local targets = worldsizes[Map.GetWorldSize()];
	self.xenomass_min, self.xenomass_max = targets[1], targets[2];
	self.firaxite_min, self.firaxite_max = targets[1], targets[2];
	self.floatstone_min, self.floatstone_max = targets[1], targets[2];
end
------------------------------------------------------------------------------
function AssignStartingPlots:PlaceKeyStrategics()
	-- This is a BE function to regulate the amounts of the "key" strategics that get placed.
	-- This regulation pertains to Float Stone, Xenomass and Firaxite.
	self:GetKeyStrategicsTargetValues()
	
	-- Process large deposits inside wild areas.
	local rand1, rand2, rand3 = self.xenomass_max - self.xenomass_min + 1, self.firaxite_max - self.firaxite_min + 1, self.floatstone_max - self.floatstone_min + 1;
	local large_xenomass_target = Map.Rand(rand1, "Number of Large Xenomass deposits to place - Lua") + self.xenomass_min;
	local large_firaxite_target = Map.Rand(rand2, "Number of Large Firaxite deposits to place - Lua") + self.firaxite_min;
	local large_floatstone_target = Map.Rand(rand3, "Number of Large Float Stone deposits to place - Lua") + self.floatstone_min;
	
	local resources_to_place = {
	{self.xenomass_ID, self.xenomass_base, self.xenomass_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.large_xenomass_list, resources_to_place, true, large_xenomass_target)
	
	local resources_to_place = {
	{self.firaxite_ID, self.firaxite_base, self.firaxite_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.softened_ice_plots, resources_to_place, true, large_firaxite_target)
	
	local resources_to_place = {
	{self.floatstone_ID, self.floatstone_base, self.floatstone_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.large_floatstone_list, resources_to_place, true, large_floatstone_target)

	-- Process small deposits inside wild areas.
	local rand1, rand2, rand3 = self.xenomass_max - self.xenomass_min + 1, self.firaxite_max - self.firaxite_min + 1, self.floatstone_max - self.floatstone_min + 1;
	local small_xenomass_target = Map.Rand(rand1, "Number of Large Xenomass deposits to place - Lua") + self.xenomass_min;
	local small_firaxite_target = Map.Rand(rand2, "Number of Large Firaxite deposits to place - Lua") + self.firaxite_min;
	local small_floatstone_target = Map.Rand(rand3, "Number of Large Float Stone deposits to place - Lua") + self.floatstone_min;
	
	local resources_to_place = {
	{self.xenomass_ID, self.minor_xenomass_base, self.minor_xenomass_range, 100, 1, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.small_xenomass_list, resources_to_place, true, small_xenomass_target)
	
	local resources_to_place = {
	{self.firaxite_ID, self.minor_firaxite_base, self.minor_firaxite_range, 100, 1, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.small_firaxite_list, resources_to_place, true, small_firaxite_target)
	
	local resources_to_place = {
	{self.floatstone_ID, self.minor_floatstone_base, self.minor_floatstone_range, 100, 1, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.small_floatstone_list, resources_to_place, true, small_floatstone_target)
	
	-- Process small deposits outside wild areas.
	local rand1, rand2, rand3 = 2 * (self.xenomass_max - self.xenomass_min + 1), 2 * (self.firaxite_max - self.firaxite_min + 1), 2 * (self.floatstone_max - self.floatstone_min + 1);
	local loose_xenomass_target = Map.Rand(rand1, "Number of Large Xenomass deposits to place - Lua") + 2 * self.xenomass_min;
	local loose_firaxite_target = Map.Rand(rand2, "Number of Large Firaxite deposits to place - Lua") + 2 * self.firaxite_min;
	local loose_floatstone_target = Map.Rand(rand3, "Number of Large Float Stone deposits to place - Lua") + 2 * self.floatstone_min;
	
	local resources_to_place = {
	{self.xenomass_ID, self.minor_xenomass_base, self.minor_xenomass_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.loose_xenomass_list, resources_to_place, true, loose_xenomass_target)
	
	local resources_to_place = {
	{self.firaxite_ID, self.minor_firaxite_base, self.minor_firaxite_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.loose_firaxite_list, resources_to_place, true, loose_firaxite_target)
	
	local resources_to_place = {
	{self.floatstone_ID, self.minor_floatstone_base, self.minor_floatstone_range, 100, 2, 2}, };
	self:BeyondEarthProcessStrategicResourceList(3, self.loose_floatstone_list, resources_to_place, true, loose_floatstone_target)
	
end
------------------------------------------------------------------------------
function AssignStartingPlots:PlaceBasicResources()
	-- KEY: {Resource ID, Quantity (0 = unquantified), weighting, minimum radius, maximum radius}
	-- KEY: (frequency (1 per n plots in the list), impact list number, plot list, resource data)
	--
	-- The radius creates a zone around the plot that other resources of that
	-- type will avoid if possible. See ProcessResourceList for impact numbers.
	--
	-- Order of placement matters, so changing the order may affect a later dependency.

	-- Basic Resources, placed singly or in groups per plot list.
	local resources_to_place = {
	{self.tubers_ID, 1, 30, 0, 1},
	{self.fiber_ID, 1, 20, 0, 1},
	{self.fruit_ID, 1, 30, 1, 1},
	{self.basalt_ID, 1, 20, 0, 1} };
	self:ProcessResourceList(11 * self.bonus_multiplier, 3, self.tundra_flat_no_feature, resources_to_place)

	local resources_to_place = {
	{self.fungus_ID, 1, 25, 0, 0},
	{self.tubers_ID, 1, 25, 0, 0},
	{self.chitin_ID, 1, 30, 0, 0},
	{self.resilin_ID, 1, 20, 0, 0} };
	self:ProcessResourceList(3 * self.bonus_multiplier, 3, self.softened_ice_plots, resources_to_place)	

	local resources_to_place = {
	{self.silica_ID, 1, 25, 0, 1},
	{self.gold_ID, 1, 30, 0, 1},
	{self.copper_ID, 1, 45, 0, 1} };
	self:ProcessResourceList(19 * self.bonus_multiplier, 3, self.next_to_softened_ice_plots, resources_to_place)	

	local resources_to_place = {
	{self.floatstone_ID, self.minor_floatstone_base, self.minor_floatstone_range, 100, 3, 4}, };
	self:BeyondEarthProcessStrategicResourceList(23, self.next_to_softened_ice_plots, resources_to_place, false, -1)

	local resources_to_place = {
	{self.basalt_ID, 1, 100, 1, 1} };
	self:ProcessResourceList(29 * self.bonus_multiplier, 3, self.dry_grass_flat_no_feature, resources_to_place)

	local resources_to_place = {
	{self.tubers_ID, 1, 100, 1, 1} };
	self:ProcessResourceList(15 * self.bonus_multiplier, 3, self.extra_deer_list, resources_to_place)

	local resources_to_place = {
	{self.chitin_ID, 1, 50, 0, 1},
	{self.fungus_ID, 1, 10, 0, 1},
	{self.fiber_ID, 1, 20, 0, 1},
	{self.fruit_ID, 1, 20, 1, 1} };
	self:ProcessResourceList(10 * self.bonus_multiplier, 3, self.desert_wheat_list, resources_to_place)

	local resources_to_place = {
	{self.tubers_ID, 1, 40, 0, 1},
	{self.basalt_ID, 1, 10, 0, 1},
	{self.silica_ID, 1, 20, 1, 1},
	{self.chitin_ID, 1, 30, 0, 1} };
	self:ProcessResourceList(10 * self.bonus_multiplier, 3, self.desert_flat_no_feature, resources_to_place)

	local resources_to_place = {
	{self.fungus_ID, 1, 50, 1, 1},
	{self.resilin_ID, 1, 50, 1, 1} };
	self:ProcessResourceList(13 * self.bonus_multiplier, 3, self.marsh_list, resources_to_place)

	local resources_to_place = {
	{self.fiber_ID, 1, 50, 1, 1},
	{self.silica_ID, 1, 30, 1, 1},
	{self.fungus_ID, 1, 20, 1, 1} };
	self:ProcessResourceList(22 * self.bonus_multiplier, 3, self.river_list, resources_to_place)

	local resources_to_place = {
	{self.silica_ID, 1, 30, 1, 1},
	{self.fungus_ID, 1, 70, 1, 1} };
	self:ProcessResourceList(22 * self.bonus_multiplier, 3, self.river_ice_plots, resources_to_place)

	local resources_to_place = {
	{self.copper_ID, 1, 30, 1, 1},
	{self.silica_ID, 1, 50, 1, 1},
	{self.chitin_ID, 1, 20, 1, 1} };
	self:ProcessResourceList(25 * self.bonus_multiplier, 3, self.plains_flat_no_feature, resources_to_place)

	local resources_to_place = {
	{self.fruit_ID, 1, 10, 2, 2},
	{self.basalt_ID, 1, 10, 1, 1},
	{self.fungus_ID, 1, 10, 1, 1},
	{self.fiber_ID, 1, 40, 0, 2},
	{self.chitin_ID, 1, 30, 1, 2} };
	self:ProcessResourceList(19 * self.bonus_multiplier, 3, self.dry_grass_flat_no_feature, resources_to_place)	

	local resources_to_place = {
	{self.fungus_ID, 1, 35, 0, 1},
	{self.gold_ID, 1, 25, 2, 3},
	{self.copper_ID, 1, 40, 1, 3} };
	self:ProcessResourceList(15 * self.bonus_multiplier, 3, self.hills_open_list, resources_to_place)	

	local resources_to_place = {
	{self.fruit_ID, 1, 65, 0, 1},
	{self.fungus_ID, 1, 35, 0, 1} };
	self:ProcessResourceList(12 * self.bonus_multiplier, 3, self.wild_area_flatlands_list, resources_to_place) 	-- Forest Wilds only.

	local resources_to_place = {
	{self.resilin_ID, 1, 100, 0, 1}, };
	self:ProcessResourceList(21 * self.bonus_multiplier, 3, self.wild_area_list, resources_to_place)		-- Forest Wilds only with this list.

	local resources_to_place = {
	{self.tubers_ID, 1, 60, 1, 2},
	{self.resilin_ID, 1, 40, 0, 1}, };
	self:ProcessResourceList(14 * self.bonus_multiplier, 3, self.wild_tundra_list, resources_to_place)

	local resources_to_place = {
	{self.tubers_ID, 1, 60, 1, 2},
	{self.resilin_ID, 1, 40, 0, 1}, };
	self:ProcessResourceList(14 * self.bonus_multiplier, 3, self.wild_desert_list, resources_to_place)

	local resources_to_place = {
	{self.tubers_ID, 1, 25, 1, 1},
	{self.basalt_ID, 1, 25, 2, 3},
	{self.gold_ID, 1, 30, 2, 3},
	{self.copper_ID, 1, 20, 1, 3} };
	self:ProcessResourceList(15 * self.bonus_multiplier, 3, self.next_to_mountain_list, resources_to_place)	

	local resources_to_place = {
	{self.basalt_ID, 1, 25, 2, 3},
	{self.gold_ID, 1, 45, 2, 3},
	{self.copper_ID, 1, 30, 1, 3} };
	self:ProcessResourceList(15 * self.bonus_multiplier, 3, self.next_to_mountain_ice_plots, resources_to_place)	

	local resources_to_place = {
	{self.basalt_ID, 1, 65, 2, 2},
	{self.silica_ID, 1, 35, 1, 3} };
	self:ProcessResourceList(31 * self.bonus_multiplier, 3, self.next_to_canyon_ice_plots, resources_to_place)	

	local resources_to_place = {
	{self.fungus_ID, 1, 25, 0, 1},
	{self.basalt_ID, 1, 45, 2, 2},
	{self.silica_ID, 1, 30, 1, 3} };
	self:ProcessResourceList(23 * self.bonus_multiplier, 3, self.next_to_canyon_list, resources_to_place)	

	local resources_to_place = {
	{self.fungus_ID, 1, 50, 2, 3},
	{self.silica_ID, 1, 50, 1, 3} };
	self:ProcessResourceList(27 * self.bonus_multiplier, 3, self.coastal_land_list, resources_to_place)

	local resources_to_place = {
	{self.resilin_ID, 1, 100, 1, 2} };
	self:ProcessResourceList(25 * self.bonus_multiplier, 3, self.next_to_wild_area_list, resources_to_place)		-- ALL Wild Areas on land covered by this list.

	local resources_to_place = {
	{self.resilin_ID, 1, 25, 1, 2},
	{self.fruit_ID, 1, 75, 2, 3} };
	self:ProcessResourceList(25 * self.bonus_multiplier, 3, self.forest_flat_that_are_not_tundra, resources_to_place)
end
------------------------------------------------------------------------------
function StartPlotSystem()
	-- Get Resources setting input by user.
	local res = Map.GetCustomOption(4)
	if res == 6 then
		res = 1 + Map.Rand(3, "Random Resources Option - Lua");
	end

	print("Creating start plot database.");
	local start_plot_database = AssignStartingPlots.Create()
	
	print("Dividing the map in to Regions.");
	-- Regional Division Method 4: Rectangle (No starts in the ice.)
	local iW, iH = Map.GetGridSize()
	local args = {
		method = 4,
		iSouthY = math.ceil(iH / 6),
		iHeight = math.ceil(iH / 3),
		resources = res,
		};
	start_plot_database:GenerateRegions(args)

	print("Choosing start locations for civilizations.");
	start_plot_database:ChooseLocations()
	
	print("Normalizing start locations and assigning them to Players.");
	start_plot_database:BalanceAndAssign()

	print("Placing Resources and City States.");
	start_plot_database:PlaceResourcesAndCityStates()

	if userInputLandmass == 3 then -- Small Continents, so
		-- tell the AI that we should treat this as a offshore expansion map
		Map.ChangeAIMapHint(4);
	elseif userInputLandmass == 4 then -- Small Continents or Islands map, so
		-- tell the AI that we should treat this as a offshore expansion map with a naval bias
		Map.ChangeAIMapHint(5);
	end

end
------------------------------------------------------------------------------
